/*  
 * Seward's algorithm d_copyEQ_u12.c modified in order to compute the
 * standard sa (not the cyclic one), to allow 0's in the input
 * and to use the small-bucket cache
 * 
 * ver 1.0 (12-apr-00)  
 * ver 1.1 (26-mag-00) works with 0's in the input text
 * ver 2.0 (12-nov-02) uses small-bucket cache
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* ==================================================================
   seward copy algorithm with small-bucket cache

   for i>0 cache[i] contains the small-bucket rank of the suffix 
   starting at position i-1. If t[i-1]==t[j-1] and we find 
   that t[i]==t[j], we know that t[i-1,i,...] and t[j-1,j,...] are in 
   the same small-bucket. Hence it make sense to compare cache[i] and
   cache[j] to determine their ordering.
   ================================================================== */

typedef int          Int32;
typedef unsigned int UInt32;
typedef char          Char;
typedef unsigned char UChar;
typedef unsigned char Bool;
#define True   ((Bool)1)
#define False  ((Bool)0)
#define N_OVERSHOOT (16-1)
#define BIGFREQ(b) (ftab[((b)+1) << 8] - ftab[(b) << 8])
#define SETMASK (1 << 30)
#define CLEARMASK (~(SETMASK))


extern int Suftest_verbose;
extern int Suftest_reverse_order;

Int32  nBlock;             // size of input string 
UChar  *block;             // input string+ overshoot
Int32  *ptr;               // suffix array
UChar  *cache;             // 8-bit cache
Int32  ftab [65537];      
Int32  runningOrder[256];

void suffixsort(UChar *x, int *p, int n)
{
  static void d_copyEQ_u12c( void );
  // ----- init global variables -----
  block=x;
  ptr=p;
  nBlock=n;
  // ---- alloc and init cache -----  
  cache = (UChar *) calloc(n+N_OVERSHOOT,1);   // set cache to zero
  // ---- do the sorting
  d_copyEQ_u12c();  
  free(cache);
}


static 
__inline__
Int32 d_copyEQ_u14_cmp0c( Int32 b1, Int32 b2 )
{
   UChar c1, c2, v1, v2;
   if (b1 == b2) return 0;

#if 0
   // 1
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 2
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 
#else
   b1 += 2; b2 += 2;  // the first two chars are always equal
#endif

   /* --- start with 7 comparisons without looking at the cache --- */ 

   // 3
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 4
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 5
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 6
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 7
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 8
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 9
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

#if 0
   // 10
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 11
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 12
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 13
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 14
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 15
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 16
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 
#endif

   //  -- check if we reached the end of string 
   if(!( b1<nBlock && b2<nBlock))
     return b2-b1;


   // execute blocks of 14 comparisons untill a difference
   // is found or we run out of the string 
   // now also look at the cache
   do {
     // 3
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 4
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 5
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 6
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 7
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 8
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 
   
     // 9
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 10
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 11
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 12
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 13
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 14
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 15
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // 16
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     v1 = cache[b1]; v2 = cache[b2];
     if (v1 != v2) return ((UInt32)v1 - (UInt32)v2);
     b1++; b2++; 

     // if (b1 >= nBlock) b1 -= nBlock;
     // if (b2 >= nBlock) b2 -= nBlock;
   } while( b1<nBlock && b2<nBlock);

   return b2-b1;   // we have  b2>b1 <=> suf2<suf1
} 


#define  TEMPLATE  d_copyEQ_u12c
#define  CMP       d_copyEQ_u14_cmp0c
#define  FMAP      ptr
#include "qsort3.template"

static
void calc_running_order ( void )
{
   Int32 i, j;
   for (i = 0; i <= 255; i++) runningOrder[i] = i;

   {
      Int32 vv;
      Int32 h = 1;
      do h = 3 * h + 1; while (h <= 256);
      do {
         h = h / 3;
         for (i = h; i <= 255; i++) {
            vv = runningOrder[i];
            j = i;
            while ( BIGFREQ(runningOrder[j-h]) > BIGFREQ(vv) ) {
               runningOrder[j] = runningOrder[j-h];
               j = j - h;
               if (j <= (h - 1)) goto zero;
            }
            zero:
            runningOrder[j] = vv;
         }
      } while (h != 1);
   }
   if(Suftest_reverse_order) {
     Int32 t;
     for(i=0;i<128;i++) {
       t=runningOrder[255-i];
       runningOrder[255-i]=runningOrder[i];
       runningOrder[i]=t;
     }
   }
}


static
void d_copyEQ_u12c ( void )
{
  // void ssort1main(UChar *, int *, int);
   static void update_sb_cache(Int32 lo, Int32 hi);
   Int32  i, j, k, ss, sb;
   UChar  c1, c2;
   Bool   bigDone[256];
   Int32  copyStart[256];
   Int32  copyEnd  [256];
   Int32  numQSorted = 0;

   for (i = 0; i < N_OVERSHOOT; i++) // init tail
      block[nBlock + i] = 0;

   for (i = 0; i <= 65536; i++) ftab[i] = 0;

   c1 = block[0];
   for (i = 1; i <= nBlock; i++) {
      c2 = block[i];
      ftab[(c1 << 8) + c2]++;
      c1 = c2;
   } 

   for (i = 1; i <= 65536; i++) ftab[i] += ftab[i-1];

   c1 = block[0];
   for (i = 0; i < nBlock; i++) {
      c2 = block[i+1];
      j = (c1 << 8) + c2;
      c1 = c2;
      ftab[j]--;
      ptr[ftab[j]] = i;
   }

   /* decide on the running order */
   calc_running_order();
   for (i = 0; i < 256; i++) bigDone[i] = False;

   /* Really do the sorting */
   for (i = 0; i <= 255; i++) {
  
      /*--
         Process big buckets, starting with the least full.
      --*/
      ss = runningOrder[i];

      /*--
         Complete the big bucket [ss] by quicksorting
         any unsorted small buckets [ss, j].  Hopefully
         previous pointer-scanning phases have already
         completed many of the small buckets [ss, j], so
         we don't have to sort them at all.
      --*/
      for (j = 0; j <= 255; j++) {
         if (j != ss) {
	   sb = (ss << 8) + j;
            if ( ! (ftab[sb] & SETMASK) ) {
               Int32 lo = ftab[sb]   & CLEARMASK;
               Int32 hi = (ftab[sb+1] & CLEARMASK) - 1;
               if (hi > lo) {
                  if (Suftest_verbose>1)
                     fprintf ( stderr,
                               "        qsort [0x%x, 0x%x]   done %d"
                               "   this %d\n",
                               ss, j, numQSorted, hi - lo + 1 );
                  d_copyEQ_u12c_qsort3 ( lo, hi );
                  numQSorted += ( hi - lo + 1 );
		  // --- update cache for sb -------
		  update_sb_cache(lo, hi);  
	         }
	       }
	       ftab[sb] |= SETMASK;
	    }
	 }

      assert (!bigDone[ss]);

      {
         for (j = 0; j <= 255; j++) {
            copyStart[j] =  ftab[(j << 8) + ss]     & CLEARMASK;
            copyEnd  [j] = (ftab[(j << 8) + ss + 1] & CLEARMASK) - 1;
         }
	 // take care of the virtual -1 char in position nBlock+1
	 if(ss==0) {
	   k=nBlock-1;
	   c1 = block[k];
	   if (!bigDone[c1])
	     ptr[ copyStart[c1]++ ] = k;
	 }
         for (j = ftab[ss << 8] & CLEARMASK; j < copyStart[ss]; j++) {
            k = ptr[j]-1; if (k < 0) continue;   // k += nBlock;
            c1 = block[k];
            if (!bigDone[c1])
               ptr[ copyStart[c1]++ ] = k;
         }
         for (j = (ftab[(ss+1) << 8] & CLEARMASK) - 1; j > copyEnd[ss]; j--) {
            k = ptr[j]-1; if (k < 0) continue; // k += nBlock;
            c1 = block[k];
            if (!bigDone[c1]) 
               ptr[ copyEnd[c1]-- ] = k;
         }
      }

      assert (copyStart[ss] - 1 == copyEnd[ss]);

      // update ftab and cache for the "copied" small buckets
      for (j = 0; j <= 255; j++) {
	sb = (j << 8) + ss; 
	if ( ! (ftab[sb] & SETMASK) ) {
	  Int32 lo = ftab[sb]   & CLEARMASK;
	  Int32 hi = (ftab[sb+1] & CLEARMASK) - 1;
	  update_sb_cache(lo, hi);
	  ftab[sb] |= SETMASK;
	}
      }
      bigDone[ss] = True;

   }

   if (Suftest_verbose>0)
   fprintf ( stderr, "        %d pointers, %d sorted, %d scanned\n",
                     nBlock, numQSorted, nBlock - numQSorted );

}

static
void update_sb_cache(Int32 lo, Int32 hi)  
{
  Int32 j, sbSize = hi-lo+1;
  //Int32 bbStart  = ftab[ss << 8] & CLEARMASK;
  //Int32 bbSize   = (ftab[(ss+1) << 8] & CLEARMASK) - bbStart;
  Int32 shifts   = 0;

  while (( (sbSize-1) >> shifts) > 255) shifts++;
  for (j = 0; j < sbSize; j++) {
    Int32 a2update     = ptr[lo + j];
    UChar qVal         = (UChar)(j >> shifts);

    assert( (j>>shifts)  < 256);
    cache[a2update+1]  = qVal;
  }
}











